use itertools::Itertools;
use tracing::debug;

use ubisync_lib::types::Peer;
use ubisync_lib::types::{ContentUpdateStrategy, PeerId};

use ubisync_lib::messages::{Message, MessageContent};

use crate::comm::conflict_resolution::merge_element_contents;
use crate::state::CommState;

pub fn handle(state: &CommState, peer: &PeerId, message: Message) {
    debug!(
        "Received message.\nFrom: {:?} (dest: {:?})\nTo: {:?} (dest: {:?})\nMessage: {message:?}",
        peer.addr(),
        peer,
        state.own_peer_id().unwrap().addr(),
        state.own_peer_id().unwrap()
    );
    match message.content() {
        MessageContent::Hello { peer_name, family } => {
            state
                .set_peer(Peer::new(
                    peer.to_owned(),
                    Some(peer_name.to_string()),
                    family.to_owned(),
                ))
                .expect("State failed");
        }
        MessageContent::JoinFamily => state.request_family_join(peer.to_owned()),
        MessageContent::AddedToFamily {
            family,
            metastate_share,
        } => {
            if state.has_family_join_request(peer.to_owned()) {
                state
                    .set_own_family(family.to_owned())
                    .expect("State failed");
                state.set_metastate_share(metastate_share.to_owned());
                debug!("Joined family, requesting metastate");
                state.send_to_peers(
                    MessageContent::Get {
                        shares: vec![metastate_share.to_owned()],
                    },
                    vec![peer.to_owned()],
                )
            }
        }
        MessageContent::LeaveFamily => state.remove_peer_from_family(peer.to_owned()),
        MessageContent::CreateElement { id, content, pot } => {
            state
                .add_received_element(
                    id.to_owned(),
                    content.to_owned(),
                    ContentUpdateStrategy::Overwrite,
                    Some(message.id().to_owned()),
                    pot.to_owned(),
                )
                .expect("State failed");
        }
        MessageContent::UpdateElement { id, content } => {
            if let Some(new_content) = merge_element_contents(
                state,
                id.clone(),
                content.to_owned(),
                message.id().to_owned(),
            ) {
                state
                    .update_element_content(
                        id.to_owned(),
                        new_content.to_owned(),
                        message.id().to_owned(),
                    )
                    .expect("State failed");
            }
        }
        MessageContent::RemoveElement { id } => {
            state.remove_element(id.to_owned()).expect("State failed");
        }
        MessageContent::AddPot { id, app_type } => {
            state
                .add_pot(id.to_owned(), app_type.to_string())
                .expect("State failed");
        }
        MessageContent::Get { shares } => {
            let _ = shares
                .iter()
                .map(|share| state.elements_of_share(share.to_owned()))
                .concat()
                .iter()
                .map(|element_id| state.get_element(element_id.to_owned()))
                .flatten()
                .map(|element| {
                    state.send_to_peers(
                        MessageContent::UpdateElement {
                            id: element.id.to_owned(),
                            content: element.content.to_owned(),
                        },
                        vec![peer.to_owned()],
                    )
                });
        },
        MessageContent::UpdateShare { share } => {
            state.set_share(share.to_owned())
        }
    }
}
