use std::collections::HashMap;

use anyhow::anyhow;
use bonsaidb::{
    core::schema::{Collection, SerializedCollection},
    local::Database,
};
use itertools::Itertools;
use serde::{Deserialize, Serialize};
use ubisync_lib::types::{FamilyId, Share, ShareContent, ShareId, SharePermissions};

use crate::state::database::{as_key::AsKey, StateDB};

#[derive(Debug, Serialize, Deserialize, Collection, PartialEq, Clone)]
#[collection(name = "shares", views = [])]
pub(super) struct DbShare {
    #[natural_id]
    pub(super) id: AsKey<ShareId>,
    pub(super) content: ShareContent,
    pub(super) members: HashMap<FamilyId, SharePermissions>,
}

impl From<DbShare> for Share {
    fn from(value: DbShare) -> Self {
        Share {
            id: (*value.id).clone(),
            content: value.content,
            members: value.members,
        }
    }
}

impl From<Share> for DbShare {
    fn from(value: Share) -> Self {
        DbShare {
            id: AsKey::new(value.id),
            content: value.content,
            members: value.members,
        }
    }
}

pub(crate) struct Shares<'a> {
    parent: &'a StateDB,
    db: &'a Database,
}

impl<'a> Shares<'a> {
    pub const fn new(parent: &'a StateDB, bonsai: &'a Database) -> Self {
        Self { parent, db: bonsai }
    }
    pub fn add(&self, share: Share) -> anyhow::Result<()> {
        DbShare::push(share.into(), self.db)
            .map(|_| ())
            .map_err(|e| anyhow!(e))
    }

    pub fn get(&self, id: ShareId) -> anyhow::Result<Option<Share>> {
        DbShare::get(&AsKey::new(id), self.db)
            .map(|share_opt| share_opt.map(|share| share.contents.into()))
            .map_err(|e| anyhow!(e))
    }

    pub fn get_all(&self) -> anyhow::Result<Vec<Share>> {
        Ok(DbShare::all(self.db)
            .query()?
            .iter()
            .map(|doc| doc.contents.clone().into())
            .collect_vec())
    }

    pub fn set(&self, share: Share) -> anyhow::Result<()> {
        match DbShare::get(&AsKey::new(share.id.clone()), self.db).map_err(|e| anyhow!(e))? {
            Some(mut doc) => doc
                .modify(self.db, |d| {
                    d.contents = DbShare {
                        id: AsKey::new(share.id.clone()),
                        content: share.content.clone(),
                        members: share.members.clone(),
                    }
                })
                .map_err(|e| anyhow!(e)),
            None => self.add(share),
        }
    }
}

#[cfg(test)]
mod tests {

    use ubisync_lib::types::{Share, ShareContent, ShareId};

    use crate::state::database::StateDB;

    #[test]
    fn add_get() {
        let db = StateDB::init(None);
        let share = Share {
            id: ShareId::new(),
            content: ShareContent { pots: vec![] },
            members: Default::default(),
        };
        db.shares().add(share.clone()).unwrap();

        let retrieved_share = db.shares().get(share.id.clone()).unwrap();
        assert_eq!(retrieved_share, Some(share))
    }
}
